"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Codec = exports.Headers = void 0;
const NoAvailableCodecError_1 = require("../../errors/NoAvailableCodecError");
const FlexibleObject_1 = require("./lib/FlexibleObject");
const ProcessPacketPostHandlers_1 = require("./lib/ProcessPacketPostHandlers");
const ARP_1 = require("./headers/ARP");
const EthernetII_1 = require("./headers/EthernetII");
const Goose_1 = require("./headers/Goose");
const IEC61850SampledValues_1 = require("./headers/IEC61850SampledValues");
const IPv4_1 = require("./headers/IPv4");
const RawData_1 = require("./headers/RawData");
const TCP_1 = require("./headers/TCP");
const UDP_1 = require("./headers/UDP");
const VLAN_802dot1Q_1 = require("./headers/VLAN_802dot1Q");
exports.Headers = [
    ARP_1.ARP,
    EthernetII_1.EthernetII,
    Goose_1.Goose,
    IEC61850SampledValues_1.IEC61850SampledValues,
    IPv4_1.IPv4,
    RawData_1.RawData,
    TCP_1.TCP,
    UDP_1.UDP,
    VLAN_802dot1Q_1.VLAN_802dot1Q
];
class Codec {
    #codecModuleConstructors = [];
    #codecSchemas = [];
    get HEADER_CODECS() {
        return this.#codecModuleConstructors;
    }
    get CODEC_SCHEMAS() {
        return this.#codecSchemas;
    }
    constructor(codecModuleConstructors = []) {
        this.#codecModuleConstructors = this.loadHeaderCodecs();
        if (codecModuleConstructors) {
            const replaced = [];
            codecModuleConstructors.forEach((codecModuleConstructor) => {
                const id = codecModuleConstructor.PROTOCOL_ID;
                this.#codecModuleConstructors.forEach((HEADER_CODEC, index, array) => {
                    if (HEADER_CODEC.PROTOCOL_ID === id) {
                        array[index] = codecModuleConstructor;
                        replaced.push(codecModuleConstructor);
                    }
                });
            });
            codecModuleConstructors
                .filter((codecModuleConstructor) => !replaced.includes(codecModuleConstructor))
                .forEach((codecModuleConstructor) => this.#codecModuleConstructors.push(codecModuleConstructor));
        }
        this.#codecSchemas = this.loadCodecSchemas();
    }
    /**
     * Load codec schemas
     * @protected
     */
    loadCodecSchemas() {
        return this.HEADER_CODECS.map((codecModuleConstructor) => ({
            id: codecModuleConstructor.PROTOCOL_ID,
            name: codecModuleConstructor.PROTOCOL_NAME,
            schema: codecModuleConstructor.PROTOCOL_SCHEMA
        }));
    }
    /**
     * Load header codecs
     * @protected
     */
    loadHeaderCodecs() {
        let headerCodecs = exports.Headers;
        headerCodecs = headerCodecs.filter((codec) => codec.PROTOCOL_ID !== RawData_1.RawData.PROTOCOL_ID);
        //Ensure RawData codec is in the end of header codecs
        headerCodecs.push(RawData_1.RawData);
        return headerCodecs;
    }
    /**
     * Internal encode headers to packet
     * @param inputs
     * @param errors
     * @private
     */
    async #encode(inputs, errors = []) {
        const codecData = {
            packet: Buffer.from([]),
            startPos: 0,
            postHandlers: []
        };
        const codecModules = [];
        for (const input of inputs) {
            const codecModuleConstructor = this.HEADER_CODECS.find((codec) => codec.PROTOCOL_ID === input.id);
            if (!codecModuleConstructor)
                continue;
            const codecModule = codecModuleConstructor.CREATE_INSTANCE(codecData, codecModules);
            codecModule.instance = new FlexibleObject_1.FlexibleObject(codecModule.validate(input.data));
            await codecModule.encode();
            codecModule.errors.forEach((errorInfo) => errors.push(errorInfo));
            codecData.startPos = codecModule.endPos;
            codecModules.push(codecModule);
        }
        return codecData;
    }
    /**
     * Internal decode packet
     * @param codecData
     * @param codecModules
     * @private
     */
    async #decode(codecData, codecModules = []) {
        const codecModuleConstructor = this.HEADER_CODECS.find((codecModuleConstructor) => codecModuleConstructor.MATCH(codecModules));
        //This unavailable error should not be thrown, the raw data codec will always match unknown data successfully
        if (!codecModuleConstructor)
            throw new NoAvailableCodecError_1.NoAvailableCodecError('No available codec constructor');
        const codecModule = codecModuleConstructor.CREATE_INSTANCE(codecData, codecModules);
        await codecModule.decode();
        codecData.startPos = codecModule.endPos;
        codecModules.push(codecModule);
        if (codecData.startPos >= codecData.packet.length)
            return;
        return this.#decode(codecData, codecModules);
    }
    /**
     * Decode packet
     * @param packet
     */
    async decode(packet) {
        const codecData = {
            packet: packet,
            startPos: 0,
            postHandlers: []
        };
        const codecModules = [];
        await this.#decode(codecData, codecModules);
        const postDecodeHandlers = (0, ProcessPacketPostHandlers_1.ProcessPacketDecodePostHandlers)(codecData.postHandlers);
        let postDecodeHandler = postDecodeHandlers.shift();
        while (postDecodeHandler) {
            await postDecodeHandler.handler();
            postDecodeHandler = postDecodeHandlers.shift();
        }
        return codecModules.map((codecModule) => ({
            id: codecModule.id,
            name: codecModule.name,
            nickname: codecModule.nickname,
            protocol: codecModule.isProtocol,
            errors: codecModule.errors,
            data: codecModule.instance.getValue()
        }));
    }
    /**
     * Encode packet
     * @param inputs
     */
    async encode(inputs) {
        const errors = [];
        const codecData = await this.#encode(inputs, errors);
        const postEncodeHandlers = (0, ProcessPacketPostHandlers_1.ProcessPacketEncodePostHandlers)(codecData.postHandlers);
        let postEncodeHandler = postEncodeHandlers.shift();
        while (postEncodeHandler) {
            await postEncodeHandler.handler();
            postEncodeHandler = postEncodeHandlers.shift();
        }
        return {
            packet: codecData.packet,
            errors: errors
        };
    }
}
exports.Codec = Codec;
